CREATE OR REPLACE FUNCTION deny_updates() RETURNS trigger AS
$$

	BEGIN { strict->import(); }

	# This should be a row level trigger */
	die 'should be called as a row level trigger' unless ($_TD->{level} eq 'ROW');

	# Do nothing if event is not UPDATE 
	return if ($_TD->{event} ne 'UPDATE');

	die "$_TD->{name}, at least the value of 'allowed' flag shoud be passed"
	unless ($_TD->{argc} >= 1);

	my $allow_arg = lc(shift @{$_TD->{args}});

	my (@deny_columns, @column_names);
	my $oldtuple = $_TD->{old};
	my $newtuple = $_TD->{new};
	
	# Convert allow_arg from PostgreSQL boolean value to either 0 or 1
	if ($allow_arg eq 'true' or $allow_arg eq 't' or 
		$allow_arg eq 'yes' or $allow_arg eq 'y' or $allow_arg eq 1) 
	{
		$allow_arg = 1;
	} elsif ($allow_arg eq 'false' or $allow_arg eq 'f' or $allow_arg eq 'no' 
			   or $allow_arg eq 'n' or $allow_arg eq 0) 
	{
		$allow_arg = 0;
	} else {
		elog(ERROR, "'allowed' param should have a boolean value");
	}

	if ($_TD->{argc} > 1) {
		@column_names = @{$_TD->{args}};
	} else {
		@column_names = ();
	}

	# check if all passed columns really exist 
	foreach (@column_names) {
		die "column '$_' doesn't exist in relation $_TD->{table_name}" 
			unless exists $oldtuple->{$_};
	}

	# Get the resulting set of columns. If include is false then we should use
	# only those columns not included in the list passed as a first param.

	if ($allow_arg == 0) {
		@deny_columns = @column_names;
	} else {

		# Make a hash of column names to speed up searching
		my %column_names_hash;

		foreach my $colname (@column_names) {
			$column_names_hash{$colname} = 1;
		}

		# Get all relation columns that are not in column_names list and put their 
		# names to the deny list.
		foreach my $colname (keys %{$oldtuple}) {
			push @deny_columns, $colname unless defined $column_names_hash{$colname};
		}
	}

	# Disallow updates for each column from deny list
	foreach (@deny_columns) {
		($oldtuple->{$_} eq $newtuple->{$_}) or 
		die "update of attribute '$_' denied by the trigger $_TD->{name}";
	}

	return;
$$ LANGUAGE plperl;

